@@ -6,11 +6,21 @@ static void Reader_dealloc(hiredis_ReaderObject *self);
 static int Reader_init(hiredis_ReaderObject *self, PyObject *args, PyObject *kwds);
 static PyObject *Reader_new(PyTypeObject *type, PyObject *args, PyObject *kwds);
 static PyObject *Reader_feed(hiredis_ReaderObject *self, PyObject *args);
-static PyObject *Reader_gets(hiredis_ReaderObject *self);
+static PyObject *Reader_gets(hiredis_ReaderObject *self, PyObject *args);
+static PyObject *Reader_setmaxbuf(hiredis_ReaderObject *self, PyObject *arg);
+static PyObject *Reader_getmaxbuf(hiredis_ReaderObject *self);
+static PyObject *Reader_len(hiredis_ReaderObject *self);
+static PyObject *Reader_has_data(hiredis_ReaderObject *self);
+static PyObject *Reader_set_encoding(hiredis_ReaderObject *self, PyObject *args, PyObject *kwds);
 
 static PyMethodDef hiredis_ReaderMethods[] = {
     {"feed", (PyCFunction)Reader_feed, METH_VARARGS, NULL },
-    {"gets", (PyCFunction)Reader_gets, METH_NOARGS, NULL },
+    {"gets", (PyCFunction)Reader_gets, METH_VARARGS, NULL },
+    {"setmaxbuf", (PyCFunction)Reader_setmaxbuf, METH_O, NULL },
+    {"getmaxbuf", (PyCFunction)Reader_getmaxbuf, METH_NOARGS, NULL },
+    {"len", (PyCFunction)Reader_len, METH_NOARGS, NULL },
+    {"has_data", (PyCFunction)Reader_has_data, METH_NOARGS, NULL },
+    {"set_encoding", (PyCFunction)Reader_set_encoding, METH_VARARGS | METH_KEYWORDS, NULL },
     { NULL }  /* Sentinel */
 };
 
@@ -68,28 +78,20 @@ static void *tryParentize(const redisReadTask *task, PyObject *obj) {
 static PyObject *createDecodedString(hiredis_ReaderObject *self, const char *str, size_t len) {
     PyObject *obj;
 
-    if (self->encoding == NULL) {
+    if (self->encoding == NULL || !self->shouldDecode) {
         obj = PyBytes_FromStringAndSize(str, len);
     } else {
-        obj = PyUnicode_Decode(str, len, self->encoding, NULL);
+        obj = PyUnicode_Decode(str, len, self->encoding, self->errors);
         if (obj == NULL) {
-            if (PyErr_ExceptionMatches(PyExc_ValueError)) {
-                /* Ignore encoding and simply return plain string. */
-                obj = PyBytes_FromStringAndSize(str, len);
-            } else {
-                assert(PyErr_ExceptionMatches(PyExc_LookupError));
-
-                /* Store error when this is the first. */
-                if (self->error.ptype == NULL)
-                    PyErr_Fetch(&(self->error.ptype), &(self->error.pvalue),
-                            &(self->error.ptraceback));
-
-                /* Return Py_None as placeholder to let the error bubble up and
-                 * be used when a full reply in Reader#gets(). */
-                obj = Py_None;
-                Py_INCREF(obj);
-            }
+            /* Store error when this is the first. */
+            if (self->error.ptype == NULL)
+                PyErr_Fetch(&(self->error.ptype), &(self->error.pvalue),
+                        &(self->error.ptraceback));
 
+            /* Return Py_None as placeholder to let the error bubble up and
+             * be used when a full reply in Reader#gets(). */
+            obj = Py_None;
+            Py_INCREF(obj);
             PyErr_Clear();
         }
     }
@@ -99,13 +101,19 @@ static PyObject *createDecodedString(hiredis_ReaderObject *self, const char *str
 }
 
 static void *createError(PyObject *errorCallable, char *errstr, size_t len) {
-    PyObject *obj;
+    PyObject *obj, *errmsg;
+
+    #if IS_PY3K
+    errmsg = PyUnicode_DecodeUTF8(errstr, len, "replace");
+    #else
+    errmsg = Py_BuildValue("s#", errstr, len);
+    #endif
+    assert(errmsg != NULL); /* TODO: properly handle OOM etc */
+
+    obj = PyObject_CallFunctionObjArgs(errorCallable, errmsg, NULL);
+    Py_DECREF(errmsg);
+    /* obj can be NULL if custom error class raised another exception */
 
-    PyObject *args = Py_BuildValue("(s#)", errstr, len);
-    assert(args != NULL); /* TODO: properly handle OOM etc */
-    obj = PyObject_CallObject(errorCallable, args);
-    assert(obj != NULL);
-    Py_DECREF(args);
     return obj;
 }
 
@@ -115,10 +123,16 @@ static void *createStringObject(const redisReadTask *task, char *str, size_t len
 
     if (task->type == REDIS_REPLY_ERROR) {
         obj = createError(self->replyErrorClass, str, len);
+        if (obj == NULL) {
+            if (self->error.ptype == NULL)
+                PyErr_Fetch(&(self->error.ptype), &(self->error.pvalue),
+                        &(self->error.ptraceback));
+            obj = Py_None;
+            Py_INCREF(obj);
+        }
     } else {
         obj = createDecodedString(self, str, len);
     }
-
     return tryParentize(task, obj);
 }
 
@@ -153,9 +167,9 @@ redisReplyObjectFunctions hiredis_ObjectFunctions = {
 };
 
 static void Reader_dealloc(hiredis_ReaderObject *self) {
+    // we don't need to free self->encoding as the buffer is managed by Python
+    // https://docs.python.org/3/c-api/arg.html#strings-and-buffers
     redisReplyReaderFree(self->reader);
-    if (self->encoding)
-        free(self->encoding);
     Py_XDECREF(self->protocolErrorClass);
     Py_XDECREF(self->replyErrorClass);
 
@@ -177,14 +191,49 @@ static int _Reader_set_exception(PyObject **target, PyObject *value) {
     return 1;
 }
 
+static int _Reader_set_encoding(hiredis_ReaderObject *self, char *encoding, char *errors) {
+    PyObject *codecs, *result;
+
+    if (encoding) {  // validate that the encoding exists, raises LookupError if not
+        codecs = PyImport_ImportModule("codecs");
+        if (!codecs)
+            return -1;
+        result = PyObject_CallMethod(codecs, "lookup", "s", encoding);
+        Py_DECREF(codecs);
+        if (!result)
+            return -1;
+        Py_DECREF(result);
+        self->encoding = encoding;
+    } else {
+        self->encoding = NULL;
+    }
+
+    if (errors) {   // validate that the error handler exists, raises LookupError if not
+        codecs = PyImport_ImportModule("codecs");
+        if (!codecs)
+            return -1;
+        result = PyObject_CallMethod(codecs, "lookup_error", "s", errors);
+        Py_DECREF(codecs);
+        if (!result)
+            return -1;
+        Py_DECREF(result);
+        self->errors = errors;
+    } else {
+        self->errors = "strict";
+    }
+
+    return 0;
+}
+
 static int Reader_init(hiredis_ReaderObject *self, PyObject *args, PyObject *kwds) {
-    static char *kwlist[] = { "protocolError", "replyError", "encoding", NULL };
+    static char *kwlist[] = { "protocolError", "replyError", "encoding", "errors", NULL };
     PyObject *protocolErrorClass = NULL;
     PyObject *replyErrorClass = NULL;
-    PyObject *encodingObj = NULL;
+    char *encoding = NULL;
+    char *errors = NULL;
 
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "|OOO", kwlist,
-        &protocolErrorClass, &replyErrorClass, &encodingObj))
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "|OOzz", kwlist,
+        &protocolErrorClass, &replyErrorClass, &encoding, &errors))
             return -1;
 
     if (protocolErrorClass)
@@ -195,28 +244,7 @@ static int Reader_init(hiredis_ReaderObject *self, PyObject *args, PyObject *kwd
         if (!_Reader_set_exception(&self->replyErrorClass, replyErrorClass))
             return -1;
 
-    if (encodingObj) {
-        PyObject *encbytes;
-        char *encstr;
-        int enclen;
-
-        if (PyUnicode_Check(encodingObj))
-            encbytes = PyUnicode_AsASCIIString(encodingObj);
-        else
-            encbytes = PyObject_Bytes(encodingObj);
-
-        if (encbytes == NULL)
-            return -1;
-
-        enclen = PyBytes_Size(encbytes);
-        encstr = PyBytes_AsString(encbytes);
-        self->encoding = (char*)malloc(enclen+1);
-        memcpy(self->encoding, encstr, enclen);
-        self->encoding[enclen] = '\0';
-        Py_DECREF(encbytes);
-    }
-
-    return 0;
+    return _Reader_set_encoding(self, encoding, errors);
 }
 
 static PyObject *Reader_new(PyTypeObject *type, PyObject *args, PyObject *kwds) {
@@ -228,6 +256,8 @@ static PyObject *Reader_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
         self->reader->privdata = self;
 
         self->encoding = NULL;
+        self->errors = "strict";  // default to "strict" to mimic Python
+        self->shouldDecode = 1;
         self->protocolErrorClass = HIREDIS_STATE->HiErr_ProtocolError;
         self->replyErrorClass = HIREDIS_STATE->HiErr_ReplyError;
         Py_INCREF(self->protocolErrorClass);
@@ -272,19 +302,26 @@ error:
     return NULL;
 }
 
-static PyObject *Reader_gets(hiredis_ReaderObject *self) {
+static PyObject *Reader_gets(hiredis_ReaderObject *self, PyObject *args) {
     PyObject *obj;
     PyObject *err;
     char *errstr;
 
+    self->shouldDecode = 1;
+    if (!PyArg_ParseTuple(args, "|i", &self->shouldDecode)) {
+        return NULL;
+    }
+
     if (redisReplyReaderGetReply(self->reader, (void**)&obj) == REDIS_ERR) {
         errstr = redisReplyReaderGetError(self->reader);
         /* protocolErrorClass might be a callable. call it, then use it's type */
         err = createError(self->protocolErrorClass, errstr, strlen(errstr));
-        obj = PyObject_Type(err);
-        PyErr_SetString(obj, errstr);
-        Py_DECREF(obj);
-        Py_DECREF(err);
+        if (err != NULL) {
+            obj = PyObject_Type(err);
+            PyErr_SetString(obj, errstr);
+            Py_DECREF(obj);
+            Py_DECREF(err);
+        }
         return NULL;
     }
 
@@ -304,3 +341,52 @@ static PyObject *Reader_gets(hiredis_ReaderObject *self) {
         return obj;
     }
 }
+
+static PyObject *Reader_setmaxbuf(hiredis_ReaderObject *self, PyObject *arg) {
+    long maxbuf;
+
+    if (arg == Py_None)
+        maxbuf = REDIS_READER_MAX_BUF;
+    else {
+        maxbuf = PyLong_AsLong(arg);
+        if (maxbuf < 0) {
+            if (!PyErr_Occurred())
+                PyErr_SetString(PyExc_ValueError,
+                                "maxbuf value out of range");
+            return NULL;
+        }
+    }
+    self->reader->maxbuf = maxbuf;
+
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
+static PyObject *Reader_getmaxbuf(hiredis_ReaderObject *self) {
+    return PyLong_FromSize_t(self->reader->maxbuf);
+}
+
+static PyObject *Reader_len(hiredis_ReaderObject *self) {
+    return PyLong_FromSize_t(self->reader->len);
+}
+
+static PyObject *Reader_has_data(hiredis_ReaderObject *self) {
+    if(self->reader->pos < self->reader->len)
+        Py_RETURN_TRUE;
+    Py_RETURN_FALSE;
+}
+
+static PyObject *Reader_set_encoding(hiredis_ReaderObject *self, PyObject *args, PyObject *kwds) {
+    static char *kwlist[] = { "encoding", "errors", NULL };
+    char *encoding = NULL;
+    char *errors = NULL;
+
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "|zz", kwlist, &encoding, &errors))
+        return NULL;
+
+    if(_Reader_set_encoding(self, encoding, errors) == -1)
+        return NULL;
+
+    Py_RETURN_NONE;
+
+}
